/*
 * $Id: CommandLineOption.java 178 2010-10-31 18:01:20Z roland $
 * Copyright (C) 2007 Roland Krueger
 * Created on 08.05.2009
 *
 * Author: Roland Krueger (www.rolandkrueger.info)
 *
 * This file is part of RoKlib.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
package info.rolandkrueger.roklib.cli;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * This class constitutes a single command line option that is managed by a
 * {@link CommandLineArgumentEvaluator}. An option has several properties that
 * can be configured through this class. An option can have the following
 * properties:<br>
 * <ul>
 * <li><b>Long option name and short option name:</b> There are two ways for the
 * user to provide an option: either as a long option or as a short option. Both
 * variants are equal. An example for that is <code>--server</code> for the long
 * option and <code>-s</code> for the short option. Both variants are configured
 * through the constructor of class {@link CommandLineOption}. It is possible to
 * omit one of these variants if, for instance, only a long option should be
 * available.
 * <li><b>Option type:</b> Each option can be of one of the following types
 * <i>flag option</i>, <i>list option</i>, or <i>singular option</i>. For more
 * information about that see {@link CommandLineOptionType}.
 * <li><b>Mandatory vs. optional options:</b> Some options can be marked as
 * mandatory. In that case, the user is compelled to provide a value for that
 * option. An example would be the target parameter for a copy command. If the
 * user forgets to provide a value for a mandatory option, the
 * {@link CommandLineArgumentEvaluator} will detect this. Note that it makes no
 * sense to mark flag options as mandatory. Therefore, trying to do so will
 * raise an unchecked exception.
 * <li><b>Default values:</b> Non-mandatory options can be provided with default
 * values. These values are set for the respective option if the user omits them
 * on the command line. To set an option's default values, you can use one of
 * the methods {@link CommandLineOption#addDefaultValue(String)} and
 * {@link CommandLineOption#addDefaultValues(String[])}.
 * </ul>
 * <h4>Creating {@link CommandLineOption} objects</h4> Since class
 * {@link CommandLineOption} is a public inner class, a special syntax has to be
 * applied in order to create new instances of this class. To do so, you need an
 * object of {@link CommandLineOption}'s enclosing class
 * {@link CommandLineArgumentEvaluator}. Say such an object is named
 * <code>evaluator</code>. A new Option can then be created with <blockquote>
 * 
 * <pre>
 *     Option option = evaluator.new Option (...);
 * </pre>
 * 
 * </blockquote> <h4>Querying an {@link CommandLineOption}'s status, i.e. the
 * user input</h4> After invoking method
 * {@link CommandLineArgumentEvaluator#evaluate()}, the status of an option can
 * be queried. That is, you can check whether the user's input is valid, which
 * values she has provided for list and singular options, and which flag options
 * she has used. To check if a flag option has been used, you can call method
 * {@link CommandLineOption#isSet()}. To obtain the list of values that have
 * been provided for a list or a singular option, method
 * {@link CommandLineOption#getInputList()} can be used. This method will return
 * a one-element list for singular options. A simple idiom for querying the
 * user-provided value of a singular option is the following: <blockquote>
 * 
 * <pre>
 * Option singularOption;
 * CommandLineArgumentEvaluator evaluator;
 * // ...
 * evaluator.evaluate ();
 * // ...
 * String value = singularOption.getInputList ()[0];
 * </pre>
 * 
 * </blockquote>
 * 
 * @see CommandLineOptionType
 * @author Roland Krueger
 */
public class CommandLineOption
{
  private String mShortOptionName;
  private String mLongOptionName;
  private String mDescription;
  private CommandLineOptionType mType;
  private boolean mIsSet = false;
  private boolean mMandatory = false;
  private boolean mDefaultValueSet = false;
  private List<String> mValues;
  private boolean mRepetionsAllowed = false;
  private int mRepetitionCount = 0;

  /**
   * List of other options which must not be provided together with this option.
   */
  private Set<CommandLineOption> mMutuallyExclusiveOptions;

  /**
   * Set of other option objects which must be provided in combination with this
   * option.
   */
  private Set<CommandLineOption> mAssociateOptions;

  private ICommandLineOptionCallback mCallback;

  /**
   * Constructor.
   * 
   * @param shortOption
   *          short version of the option name. For example, <code>-s</code>
   * @param longOption
   *          long version of the option name. For example,
   *          <code>--server_address</code>
   * @param description
   *          short description for the option. This description will be used
   *          for assembling a help message by method
   *          {@link CommandLineArgumentEvaluator#getOptionDescription()}.
   * @param type
   *          type of the option. An option can be a flag option, a singular
   *          option, or a list option.
   * @param mandatory
   *          <code>true</code> if the option is mandatory. Is only allowed for
   *          list and singular options
   * @throws IllegalArgumentException
   *           either if an option was both configured to be a flag option and
   *           mandatory or if both short and long option name were left blank
   */
  public CommandLineOption (Character shortOption, String longOption, String description,
      CommandLineOptionType type, boolean mandatory, ICommandLineOptionCallback callback)
  {
    this ();
    setCallback (callback);
    if (type == CommandLineOptionType.FLAG && mandatory == true)
      throw new IllegalArgumentException ("A flag option must not be mandatory.");
    if (shortOption == null)
      mShortOptionName = "";
    else
      mShortOptionName = String.valueOf (shortOption);
    if (longOption == null) longOption = "";
    if (mShortOptionName.equals ("") && longOption.equals (""))
      throw new IllegalArgumentException ("Short and long option must not be both empty");
    mLongOptionName = longOption;
    mType = type;
    mMandatory = mandatory;
    if (description == null) description = "";
    mDescription = description;
  }

  public CommandLineOption (Character shortOption, String longOption, String description,
      CommandLineOptionType type)
  {
    this (shortOption, longOption, description, type, false);
  }

  public CommandLineOption (Character shortOption, String longOption, String description,
      CommandLineOptionType type, boolean mandatory)
  {
    this (shortOption, longOption, description, type, mandatory, new EmptyCallback ());
  }

  /**
   * Private default constructor.
   */
  private CommandLineOption ()
  {
    mCallback = new EmptyCallback ();
    mValues = new LinkedList<String> ();
    mShortOptionName = "";
    mLongOptionName = "";
  }

  /**
   * Constructor for creating a {@link CommandLineOption} without a description.
   */
  public CommandLineOption (Character shortOption, String longOption, CommandLineOptionType type,
      boolean mandatory)
  {
    this (shortOption, longOption, "", type, mandatory);
  }

  public void setCallback (ICommandLineOptionCallback callback)
  {
    if (callback == null) throw new NullPointerException ("Callback is null");
    mCallback = callback;
  }

  public void allowOptionRepetition (boolean yesNo)
  {
    if (yesNo == true && mType == CommandLineOptionType.SINGULAR)
      throw new IllegalStateException ("Cannot allow option repetition for single value options.");

    mRepetionsAllowed = yesNo;
  }

  public void setMandatory (boolean mandatory)
  {
    if (isFlag ()) throw new IllegalStateException ("A flag option must not be mandatory.");
    mMandatory = mandatory;
  }

  public boolean isOptionRepetitionAllowed ()
  {
    return mRepetionsAllowed;
  }

  /**
   * Checks whether this option is a flag option.
   * 
   * @return <code>true</code> if this option was configured as a flag option
   */
  public boolean isFlag ()
  {
    return mType == CommandLineOptionType.FLAG;
  }

  protected void invokeCallback (CommandLineArgumentEvaluator source)
  {
    mCallback.optionIsSet (source, mValues);
  }

  /**
   * Checks whether the option was used on the command line. This is useful for
   * both flag options and non-mandatory options. Mandatory options that haven't
   * been provided by the user can be queried with
   * {@link CommandLineArgumentEvaluator#getMissingOptions()}.
   * 
   * @return <code>true</code> if the option was used by the user on the command
   *         line.
   */
  public boolean isSet ()
  {
    return mIsSet;
  }

  public void addAssociateOptions (CommandLineOption... associates)
  {
    // first check if both options which shall go together havn't earlier been
    // defined as
    // mutually exclusive
    for (CommandLineOption option : associates)
      if (isMutuallyExclusiveWith (option))
        throw new IllegalArgumentException (String.format (
            "Option %s is already defined to be mutually " + "exclusive to this option.", option));

    if (mAssociateOptions == null) mAssociateOptions = new HashSet<CommandLineOption> ();

    for (CommandLineOption option : associates)
      mAssociateOptions.add (option);
  }

  public void addMutuallyExclusiveOptions (CommandLineOption... otherOptions)
  {
    // first check if both options which shall be mutually exclusive havn't
    // earlier been defined
    // as associate options

    if (mMutuallyExclusiveOptions == null)
      mMutuallyExclusiveOptions = new HashSet<CommandLineOption> ();

    for (CommandLineOption option : otherOptions)
    {
      if (isAssociateTo (option))
        throw new IllegalArgumentException (String.format (
            "Option %s is already defined as an associate option " + "to this option.", option));

      // check if both options are mandatory. If so, they cannot be mutually
      // exclusive
      if (isMandatory () || option.isMandatory ())
        throw new IllegalArgumentException (String.format (
            "Both this option and option %s which shall be " + "mutually exclusive are mandatory.",
            option));

      // check if one of the options involved is already set (for example with a
      // default value). If so
      // adding them to a mutually exclusive relationship is disallowed
      if (isSet () || option.isSet ())
        throw new IllegalArgumentException (String.format (
            "Cannot add to mutually exclusive relationship. Either "
                + "this option, both options or option %s are already set.", option));

      mMutuallyExclusiveOptions.add (option);
      if (! option.isMutuallyExclusiveWith (this)) option.addMutuallyExclusiveOptions (this);
    }
  }

  public boolean isMutuallyExclusiveWith (CommandLineOption otherOption)
  {
    if (mMutuallyExclusiveOptions == null) return false;
    return (mMutuallyExclusiveOptions.contains (otherOption));
  }

  public boolean isAssociateTo (CommandLineOption otherOption)
  {
    if (mAssociateOptions == null) return false;
    return (mAssociateOptions.contains (otherOption));
  }

  /**
   * Returns the list of values that have been provided by the user for an
   * option. If this {@link CommandLineOption} object is a singular option, the
   * returned list has exactly one element.
   * 
   * @return the list of values that have been provided by the user for an
   *         option or <code>null</code> if there aren't any values for this
   *         option
   */
  public String[] getInputList ()
  {
    if (mValues.size () == 0) return new String[] {};
    String[] result = new String[mValues.size ()];
    mValues.toArray (result);
    return result;
  }

  public String getSingleParameterValue ()
  {
    if (mType == CommandLineOptionType.FLAG)
      throw new IllegalStateException (
          "This command line option is of type FLAG. It cannot provide a value.");
    if (! isSet ()) throw new IllegalStateException ("This option has not been set.");
    assert mValues.size () != 0;
    return mValues.get (0);
  }

  /**
   * Mark this option as set.
   */
  protected void set ()
  {
    mIsSet = true;
  }

  /**
   * Returns whether this option is mandatory or optional.
   * 
   * @return <code>true</code> if this option is marked as mandatory
   */
  public boolean isMandatory ()
  {
    return mMandatory;
  }

  protected boolean addUserInput (String input)
  {
    // check if default values have benn provided. If so, remove them since user
    // input is available
    if (mDefaultValueSet)
    {
      mValues.clear ();
      mDefaultValueSet = false;
    }
    set ();
    // has a value already been given for this option if it is singular?
    if (mType == CommandLineOptionType.SINGULAR && mValues.size () == 1) return false;
    // if this option is a flag, don't accept input
    if (mType == CommandLineOptionType.FLAG) return false;

    // otherwise accept input
    mValues.add (input);
    return true;
  }

  @Override
  public boolean equals (Object obj)
  {
    if (obj == null) return false;
    if (obj == this) return true;
    if (obj instanceof CommandLineOption)
    {
      CommandLineOption other = (CommandLineOption) obj;
      if (! mLongOptionName.equals ("") && mLongOptionName.equals (other.mLongOptionName))
        return true;
      if (! mShortOptionName.equals ("") && mShortOptionName.equals (other.mShortOptionName))
        return true;
    }
    return false;
  }

  @Override
  public int hashCode ()
  {
    return super.hashCode ();
  }

  /**
   * Returns the description of this option as provided through the constructor.
   * 
   * @return the description of this option.
   */
  public String getDescription ()
  {
    return mDescription;
  }

  /**
   * Adds a default value for this option. Every non-mandatory
   * {@link CommandLineOption} object maintains a list of default values. If the
   * user has omitted an option on the command line, querying the option with
   * {@link CommandLineOption#getInputList()} will yield its default values.
   * Additionally, each option with default values will return <code>true</code>
   * wenn invoking its {@link CommandLineOption#isSet()} method.<br>
   * <br>
   * Repeated calls to {@link CommandLineOption#addDefaultValue(String)} will
   * add each provided value to the list of default values for list options.
   * This is not the case for singular options. Here, only the value provided
   * with the last call to {@link CommandLineOption#addDefaultValue(String)}
   * will be used as the only default value.<br>
   * <br>
   * Note that default values should usually be set before evaluating the
   * command line arguments with {@link CommandLineArgumentEvaluator#evaluate()}
   * . Otherwise the actual data provided by the user will be overwritten for
   * singular options or the default values will be added to the user's data for
   * list options, respectively.
   * 
   * @param defaultValue
   *          the default value to be added
   * @throws IllegalStateException
   *           if this option is either a flag option (which aren't meant to be
   *           provided with a value by definition) or a mandatory option (where
   *           the user is forced to provide a value himself).
   */
  public void addDefaultValue (String defaultValue)
  {
    if (mType == CommandLineOptionType.FLAG)
      throw new IllegalStateException ("Cannot set default value for a flag option.");
    if (mMandatory)
      throw new IllegalStateException (
          "Setting a default value of a mandatory option is not allowed.");
    if (mMutuallyExclusiveOptions != null && mMutuallyExclusiveOptions.size () > 0)
      throw new IllegalStateException (
          "This option is already part of a mutually exclusive relationship.");

    if (mType == CommandLineOptionType.SINGULAR)
    {
      mValues = new LinkedList<String> ();
    }
    mValues.add (defaultValue);
    mDefaultValueSet = true;
    set ();
  }

  /**
   * Adds a list of values to the default values of this option. For singular
   * options, this list must not contain more than one element.
   * 
   * @see #addDefaultValue(String)
   * @param values
   *          list of default values
   * @throws UnsupportedOperationException
   *           if this option is a flag option or if the passed
   *           <code>values</code> list has more than one element for singular
   *           options
   */
  public void addDefaultValues (String... values)
  {
    if (mType == CommandLineOptionType.FLAG)
      throw new UnsupportedOperationException ("Cannot set a default value for a flag option.");
    if (values.length > 1 && mType == CommandLineOptionType.SINGULAR)
      throw new UnsupportedOperationException ("Cannot set more than one default value "
          + "for a singular option. Use setDefaultValue(String) instead.");
    for (String value : values)
    {
      addDefaultValue (value);
    }
  }

  /**
   * Returns the option name's long version.
   * 
   * @return the option name's long version
   */
  public String getLongOption ()
  {
    return mLongOptionName;
  }

  /**
   * Returns the option name's short version.
   * 
   * @return the option name's short version
   */
  public String getShortOption ()
  {
    return mShortOptionName;
  }

  public CommandLineOptionType getType ()
  {
    return mType;
  }

  public void increaseRepetitions ()
  {
    mRepetitionCount++;
  }

  public int getRepetitionCount ()
  {
    return mRepetitionCount;
  }

  @Override
  public String toString ()
  {
    StringBuilder buf = new StringBuilder ();
    buf.append ("['").append (mShortOptionName).append ("', '");
    buf.append (mLongOptionName).append ("', ").append ("'");
    buf.append (mDescription).append ("', ");
    buf.append (mType.toString ());
    buf.append (", mandatory=").append (mMandatory);
    buf.append ("]");
    return buf.toString ();
  }

  private static class EmptyCallback implements ICommandLineOptionCallback
  {
    public void optionIsSet (CommandLineArgumentEvaluator source, List<String> paramValues)
    {
    };
  }
}
