/*
 * $Id: ConsoleCommandObject.java 178 2010-10-31 18:01:20Z roland $
 * Copyright (C) 2007 Roland Krueger
 * Created on 05.01.2006
 *
 * Author: Roland Krueger (www.rolandkrueger.info)
 *
 * This file is part of RoKlib.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
package info.rolandkrueger.roklib.cli;

import java.io.OutputStream;

/**
 * Base class for all command objects that are used with the
 * {@link ConsoleThread}. The command's concrete functionality is implemented by
 * overriding the <tt>run()</tt> method. <br> {@link ConsoleCommandObject} contains
 * three strings: the command's name that is used by the user on the console, a
 * short description of the command that is used by the '<tt>help</tt>' command
 * to show a brief overview of the available commands, and a long description
 * that can be used to describe the command's functioning in more depth.
 * 
 * @author Roland Krueger
 */
public abstract class ConsoleCommandObject extends AbstractCommandObject
{
  private final String mShortDescription;
  private final String mLongDescription;
  protected OutputStream mOutputStream;

  /**
   * Constructs a new command with the specified name and descriptions.
   * 
   * @param name
   *          The name of the command.
   * @param shortDescription
   *          A short description of the command.
   * @param longDescription
   *          A more detailed description of the command's functioning.
   */
  public ConsoleCommandObject (String name, String shortDescription, String longDescription)
  {
    super (name);
    mShortDescription = shortDescription;
    mLongDescription = longDescription;
  }

  /**
   * Constructs a new command with the specified name and descriptions.
   * 
   * @param name
   *          The name of the command.
   * @param shortDescription
   *          A short description of the command.
   * @param longDescription
   *          A more detailed description of the command's functioning.
   * @param output
   *          An output stream that can be used to write the command's output
   *          to.
   */
  public ConsoleCommandObject (String name, String shortDescription, String longDescription,
      OutputStream output)
  {
    this (name, shortDescription, longDescription);
    mOutputStream = output;
  }

  /**
   * Returns the output stream for this command object. This stream can be used
   * by command objects to print out messages.
   * 
   * @return the output stream for this command object
   */
  public OutputStream getOutputStream ()
  {
    return mOutputStream;
  }

  /**
   * Sets the output stream for a command object.
   * 
   * @param stream
   *          the output stream for the command object
   */
  public void setOutputStream (OutputStream stream)
  {
    mOutputStream = stream;
  }

  /**
   * Returns the short description for this command object.
   * 
   * @return the short description for this command object.
   */
  public String getShortDescription ()
  {
    return mShortDescription;
  }

  /**
   * Returns the long description for this command object.
   * 
   * @return the long description for this command object.
   */
  public String getLongDescription ()
  {
    return mLongDescription;
  }
}
