/*
 * $Id: ConsoleThread.java 178 2010-10-31 18:01:20Z roland $
 * Copyright (C) 2007 Roland Krueger
 * Created on 04.01.2006
 *
 * Author: Roland Krueger (www.rolandkrueger.info)
 *
 * This file is part of RoKlib.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
package info.rolandkrueger.roklib.cli;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.logging.Logger;

/**
 * This is a class that provides a shell similar to those of programs like
 * <tt>gdb</tt> or <tt>bc</tt>. This class is running in a thread. It simply
 * reads one line after the other from an input stream and hands it over to a
 * given handler. This handler, which must implement the interface
 * {@link CommandHandler}, can then process the command string. The input stream
 * from which the user commands are read and the output stream to which the
 * console's output is written are configurable. If these streams are not
 * specified they default to the standard input and standard ouput. Thus, it is
 * possible to read in commands from a Swing frame, for instance.
 * 
 * @see CommandHandler
 * @see ConsoleCommandObject
 * @author Roland Krueger
 */
public class ConsoleThread extends Thread
{
  private static Logger sLogger = Logger.getLogger (ConsoleThread.class.getPackage ().getName ());

  private boolean mShutdownRequested = false;
  private CommandHandler mCommandHandler;
  private String mGreetingMsg;
  private Object mPrompt;
  private BufferedReader mReader;
  private InputStream mInputStream;
  private PrintStream mOutput;

  /**
   * Default constructor.
   */
  private ConsoleThread ()
  {
  }

  /**
   * Constructs a new command console with configurable input and output
   * streams.
   * 
   * @param handler
   *          Class that gets passed the user input to process it further.
   * @param greetingMsg
   *          A greeting message that is shown when the console starts up.
   * @param prompt
   *          The command prompt. Can be something like '$' or '>' or a mutable
   *          object. The <tt>toString()</tt> method of the prompt-object is
   *          used to determine the string to be used as command prompt. For
   *          example, use a date object here that will return the current date
   *          and time in its <tt>toString()</tt> method.
   * @param outStream
   *          Output stream that will be used to write the command prompt to.
   *          Note that the output of specific commands will not be written to
   *          this output stream if this is not configured by the respective
   *          command objects. See org.pi4.simplesim.util.CommandObject for
   *          further information.
   * @param inStream
   *          Input stream that is used to read the user's input from.
   */
  public ConsoleThread (CommandHandler handler, String greetingMsg, Object prompt,
      OutputStream outStream, InputStream inStream)
  {
    this ();
    mReader = new BufferedReader (new InputStreamReader (inStream));
    mInputStream = inStream;
    mOutput = new PrintStream (System.out);
    mCommandHandler = handler;
    mPrompt = prompt;
    mGreetingMsg = greetingMsg;
    startThread ();
  }

  /**
   * Constructs a new command console with <tt>stdin</tt> and <tt>stdout</tt> as
   * the default input and output streams.
   * 
   * @param handler
   *          Class that gets passed the user input to process it further.
   * @param greetingMsg
   *          A greeting message that is shown when the console starts up.
   * @param prompt
   *          The command prompt. Can be something like '$' or '>' or a mutable
   *          object. The <tt>toString()</tt> method of the prompt-object is
   *          used to determine the string to be used as command prompt. For
   *          example, use a date object here that will return the current date
   *          and time in its <tt>toString()</tt> method.
   */
  public ConsoleThread (CommandHandler handler, String greetingMsg, String prompt)
  {
    this (handler, greetingMsg, prompt, System.out, System.in);
  }

  /**
   * Starts the console thread.
   */
  private void startThread ()
  {
    assert mGreetingMsg != null && mPrompt != null;
    mOutput.println (mGreetingMsg);
    mOutput.print (mPrompt);
    start ();
    sLogger.info ("Console thread started.");
  }

  @Override
  public void run ()
  {
    while (! mShutdownRequested)
    {
      try
      {
        String line = mReader.readLine ();
        mCommandHandler.handleCommand (line);
        mOutput.print (mPrompt);
      } catch (IOException ioExc)
      {
        // TODO: error handling
        ioExc.printStackTrace ();
      }
    }
    sLogger.info ("Shutting down console thread.");
  }

  /**
   * Changes the console's prompt to a new prompt object. The
   * <tt>toString()</tt> method of the prompt object is used to determine the
   * string to be used as the command prompt.
   * 
   * @param newPrompt
   *          the new command line prompt to be used by this console.
   */
  public synchronized void setPrompt (Object newPrompt)
  {
    assert newPrompt != null;
    mPrompt = newPrompt;
  }

  /**
   * Returns the object that represents the console prompt. This may be a string
   * with a character sequence, such as <code>"&gt;"</code>, for example. To
   * print out the prompt on the console, the prompt object's
   * <code>toString</code> method is used.
   * 
   * @return the object that represents the console prompt
   */
  public Object getPrompt ()
  {
    return mPrompt;
  }

  /**
   * Closes the console and stops the associated thread.
   */
  public synchronized void shutdown ()
  {
    sLogger.info ("Requesting shutdown of console thread.");
    mShutdownRequested = true;
    try
    {
      mReader.close ();
    } catch (IOException ioExc)
    {
      // TODO: error handling
      ioExc.printStackTrace ();
    }
  }

  /**
   * Returns the output stream where the console's output is written to.
   * 
   * @return the output stream where the console's output is written to.
   */
  public OutputStream getOutputStream ()
  {
    return mOutput;
  }

  /**
   * Returns the input stream from which the user's input is read from.
   * 
   * @return the input stream from which the user's input is read from.
   */
  public InputStream getInputStream ()
  {
    return mInputStream;
  }

  /**
   * Sets the output stream where the console's output is written to.
   * 
   * @param outStream
   *          the output stream where the console's output is written to.
   */
  public synchronized void setOutputStream (OutputStream outStream)
  {
    assert outStream != null;
    mOutput = new PrintStream (outStream);
  }

  /**
   * Sets the input stream from which the user's input is read from.
   * 
   * @param inStream
   *          the input stream from which the user's input is read from.
   */
  public synchronized void setInputStream (InputStream inStream)
  {
    assert inStream != null;
    mReader = new BufferedReader (new InputStreamReader (inStream));
    mInputStream = inStream;
  }
}
