/*
 * $Id: ManagedValue.java 224 2011-01-05 19:18:38Z roland $
 * Copyright (C) 2007 Roland Krueger
 * Created on 11.12.2009
 *
 * Author: Roland Krueger (www.rolandkrueger.info)
 *
 * This file is part of RoKlib.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
package info.rolandkrueger.roklib.util.data;

import info.rolandkrueger.roklib.util.helper.CheckForNull;

/**
 * Wrapper class for a single data value that monitors the status of and access
 * to this value. A {@link ManagedValue} holds one piece of data that can be of
 * an arbitrary type. It provides additional services for this data that is
 * otherwise not available or only in a restricted way to the data object
 * itself.
 * <p>
 * One recurrent problem with data that does not necessarily need to be
 * initialized is for external code to determine whether the value is valid or
 * wasn't initialized yet. When the data is available as an object, one could
 * test if the object reference is <code>null</code>. This isn't possible any
 * longer in the instance that the <code>null</code> value is a valid value for
 * the object. The same applies to primitive data types. It is very clumsy if
 * not impossible to define a specific value, such as <code>0</code>, as a
 * marker for the status of the primitive value.
 * <p>
 * Class {@link ManagedValue} alleviates this problem. It encapsulates a single
 * value together with its status without reserving a special value as a status
 * label. Thus, every possible value, such as <code>null</code> or
 * <code>0</code> can be used as valid data for the encapsulated object.
 * Furthermore, {@link ManagedValue} lets you query the status of the encapsulated
 * data object. You can determine whether the object has been set, deleted or
 * changed, for instance.
 * <p>
 * Another feature of {@link ManagedValue} is that it allows the locking of the
 * underlying data object. That is, the owner of the {@link ManagedValue} can
 * hold an exclusive lock on the value, that prevents other objects from
 * changing the data of the {@link ManagedValue}.
 * 
 * @author Roland Krueger
 * @param <T>
 *          the data value that this class is supposed to manage
 */
public class ManagedValue<T>
{
  /**
   * State of the encapsulated data object. The data can be in one of four states:
   * <ul>
   * <li>{@link ManagedValue.StatusEnum#UNDEFINED} - the value hasn't been
   * initialized yet. Trying to access it will result in an
   * {@link IllegalStateException}.</li>
   * <li>{@link ManagedValue.StatusEnum#SET} - the value has been set and can be
   * queried with {@link ManagedValue#getValue()}</li>
   * <li>{@link ManagedValue.StatusEnum#DELETED} - the value has been deleted.
   * Trying to access it will result in an {@link IllegalStateException}.</li>
   * <li>{@link ManagedValue.StatusEnum#CHANGED} - the value has been changed. It
   * can still be accessed.</li>
   * </ul>
   */
  public enum StatusEnum
  {
    UNDEFINED, SET, DELETED, CHANGED
  }

  private T mValue;
  private StatusEnum mStatus;
  private Object mWriteLock;

  /**
   * Default constructor. Sets the state of the underlying data object to
   * {@link StatusEnum#UNDEFINED}.
   */
  public ManagedValue ()
  {
    mStatus = StatusEnum.UNDEFINED;
  }

  /**
   * Creates a new {@link ManagedValue} with an initial value object. The state
   * of the newly created {@link ManagedValue} is
   * {@link ManagedValue.StatusEnum#SET}.
   * 
   * @param initialValue
   *          the initial value object
   */
  public ManagedValue (T initialValue)
  {
    setValue (initialValue);
  }

  /**
   * Returns <code>true</code> if the current state of the {@link ManagedValue}
   * is either {@link ManagedValue.StatusEnum#SET} or
   * {@link ManagedValue.StatusEnum#CHANGED}.
   * 
   * @return <code>true</code> if the current state of the {@link ManagedValue}
   *         is either {@link ManagedValue.StatusEnum#SET} or
   *         {@link ManagedValue.StatusEnum#CHANGED}.
   */
  public boolean canRead ()
  {
    return isSet () || mStatus == StatusEnum.CHANGED;
  }

  /**
   * Unsets the data object. The new state of the {@link ManagedValue} is
   * {@link ManagedValue.StatusEnum#DELETED}.
   */
  public void deleteValue ()
  {
    mStatus = StatusEnum.DELETED;
  }

  /**
   * Returns the current state of the data value.
   * 
   * @return the current state of the data value.
   * @see ManagedValue.StatusEnum
   */
  public StatusEnum getState ()
  {
    return mStatus;
  }

  /**
   * Returns the value object that is managed by this {@link ManagedValue}.
   * 
   * @return the value object that is managed by this {@link ManagedValue}.
   * @throws IllegalStateException
   *           if the value either hasn't been initialized yet or was deleted
   *           with {@link ManagedValue#deleteValue()}
   */
  public T getValue ()
  {
    if (mStatus == StatusEnum.UNDEFINED)
    {
      throw new IllegalStateException ("Unable to get value: value is still undefined.");
    }
    if (mStatus == StatusEnum.DELETED)
    {
      throw new IllegalStateException ("Unable to get value: value has been deleted.");
    }

    return mValue;
  }

  /**
   * Copy constructor. Copies the exact state of the other {@link ManagedValue}
   * into this object.
   * 
   * @param other
   *          an other {@link ManagedValue}
   */
  public ManagedValue (ManagedValue<T> other)
  {
    this ();
    mValue = other.mValue;
    mStatus = other.mStatus;
    mWriteLock = other.mWriteLock;
  }

  /**
   * Locks a {@link ManagedValue}. By setting a lock, an owner of the
   * {@link ManagedValue} can assure that no other external process can alter
   * the {@link ManagedValue}'s data object or change its state. If that is
   * being attempted by an object that doesn't own the lock key, an
   * {@link IllegalStateException} will be raised. Unlocking the
   * {@link ManagedValue} is only possible by passing the key object to the
   * respective unlock methods.
   * 
   * @param writeLock
   *          an arbitrary object instance that functions as the key for this
   *          lock
   * @throws NullPointerException
   *           if the key object is <code>null</code>
   * @see ManagedValue#setValue(Object)
   * @see ManagedValue#setValue(Object, Object)
   */
  public void lock (Object writeLock)
  {
    CheckForNull.check (writeLock);
    mWriteLock = writeLock;
  }

  /**
   * Returns <code>true</code> if the {@link ManagedValue} is currently locked.
   * 
   * @return <code>true</code> if the {@link ManagedValue} is currently locked.
   */
  public boolean isLocked ()
  {
    return mWriteLock != null;
  }

  /**
   * Resets the state of the data value. This will only have any effect if the
   * current status is {@link ManagedValue.StatusEnum#CHANGED}. If that is the case
   * the state is reset to {@link ManagedValue.StatusEnum#SET}. Thus,
   * {@link ManagedValue#reset()} can be used to acknowledge that the
   * {@link ManagedValue}'s data has been changed.
   */
  public void reset ()
  {
    if (mStatus == StatusEnum.CHANGED) 
    {
      mStatus = StatusEnum.SET;
    }
  }

  /**
   * Sets the value of the data object. The {@link ManagedValue} must not be
   * locked when this method is called. The new state of the value object is
   * either {@link ManagedValue.StatusEnum#SET} if it was undefined or deleted
   * before this operation or {@link ManagedValue.StatusEnum#CHANGED} if it was
   * already set.
   * 
   * @param value
   * @throws IllegalStateException
   *           if this {@link ManagedValue} is locked
   */
  public void setValue (T value)
  {
    CheckForNull.check (value);
    if (isLocked ()) 
    {
      throw new IllegalStateException ("Value object is currently locked.");
    }
    setValueInternal (value);
  }

  /**
   * Sets the value of the data object. This method can be used if and only if
   * the {@link ManagedValue} is currently locked. The lock's key is provided as
   * the second parameter. The new state of the value object is either
   * {@link ManagedValue.StatusEnum#SET} if it was undefined or deleted before this
   * operation or {@link ManagedValue.StatusEnum#CHANGED} if it was already set.
   * 
   * @param value
   *          a new value
   * @param lockKey
   *          the object that was used to lock the {@link ManagedValue}
   * @throws IllegalStateException
   *           if the {@link ManagedValue} isn't currently locked
   * @throws IllegalArgumentException
   *           if the passed lock key is not the same object instance as the key
   *           object that was used to set the lock
   * @see ManagedValue#unlock(Object)
   */
  public void setValue (T value, Object lockKey)
  {
    if (! isLocked ())
    {
      throw new IllegalStateException ("Cannot unlock value: object is not locked.");
    }
    if (mWriteLock != lockKey)
    {
      throw new IllegalArgumentException ("Cannot set value: incorrect lock key.");
    }
    setValueInternal (value);
  }

  private void setValueInternal (T value)
  {
    mValue = value;
    if (mStatus == StatusEnum.SET || mStatus == StatusEnum.CHANGED)
    {
      mStatus = StatusEnum.CHANGED;
    } else
    {
      mStatus = StatusEnum.SET;
    }
  }

  /**
   * Returns <code>true</code> if the current state of the {@link ManagedValue}
   * is {@link ManagedValue.StatusEnum#SET}.
   * 
   * @return <code>true</code> if the current state of the {@link ManagedValue}
   *         is {@link ManagedValue.StatusEnum#SET}.
   */
  public boolean isSet ()
  {
    return mStatus == StatusEnum.SET;
  }

  /**
   * Returns <code>true</code> if the current state of the {@link ManagedValue}
   * is {@link ManagedValue.StatusEnum#CHANGED}.
   * 
   * @return <code>true</code> if the current state of the {@link ManagedValue}
   *         is {@link ManagedValue.StatusEnum#CHANGED}.
   */
  public boolean isChanged ()
  {
    return mStatus == StatusEnum.CHANGED;
  }

  /**
   * Returns <code>true</code> if the current state of the {@link ManagedValue}
   * is {@link ManagedValue.StatusEnum#UNDEFINED}.
   * 
   * @return <code>true</code> if the current state of the {@link ManagedValue}
   *         is {@link ManagedValue.StatusEnum#UNDEFINED}.
   */
  public boolean isUndefined ()
  {
    return mStatus == StatusEnum.UNDEFINED;
  }

  /**
   * Returns <code>true</code> if the current state of the {@link ManagedValue}
   * is {@link ManagedValue.StatusEnum#DELETED}.
   * 
   * @return <code>true</code> if the current state of the {@link ManagedValue}
   *         is {@link ManagedValue.StatusEnum#DELETED}.
   */
  public boolean isDeleted ()
  {
    return mStatus == StatusEnum.DELETED;
  }

  @Override
  @SuppressWarnings (value = "all")
  public boolean equals (Object obj)
  {
    if (obj == null) {return false;}
    if (obj == this) {return true;}
    if (obj instanceof ManagedValue)
    {
      ManagedValue other = (ManagedValue) obj;
      if (! (mStatus == StatusEnum.SET || mStatus == StatusEnum.CHANGED)) return false;
      if (! (other.mStatus == StatusEnum.SET || other.mStatus == StatusEnum.CHANGED)) return false;

      return mValue.equals (other.mValue);
    }
    return false;
  }

  @Override
  public int hashCode ()
  {
    if (mStatus == StatusEnum.SET || mStatus == StatusEnum.CHANGED)
    {
      return mValue.hashCode ();
    } else
    {
      return super.hashCode ();
    }
  }

  @Override
  public String toString ()
  {
    if (canRead ()) 
    {
      return mValue.toString ();
    }
    return String.format ("%s:%s", super.toString (), mStatus.toString ());
  }

  /**
   * Unlocks a previously locked {@link ManagedValue}. Unlock will only succeed
   * if the given key has the same object reference as the key object that was
   * used to lock the {@link ManagedValue} with
   * {@link ManagedValue#lock(Object)}. The given key is compared with the lock
   * key with the <code>==</code> operator.
   * 
   * @param writeLock
   *          the object that was used to lock the {@link ManagedValue}
   * @throws IllegalStateException
   *           if the {@link ManagedValue} isn't currently locked
   * @throws IllegalArgumentException
   *           if the passed lock key is not the same object instance as the key
   *           object that was used to set the lock
   * @see ManagedValue#lock(Object)
   */
  public void unlock (Object writeLock)
  {
    if (! isLocked ())
    {
      throw new IllegalStateException ("Cannot unlock value: object is not locked.");
    }
    if (mWriteLock != writeLock)
    {
      throw new IllegalArgumentException ("Cannot unlock value: wrong key.");
    }
    mWriteLock = null;
  }
}
