/*
 * $Id: ResourceHandlingManager.java 178 2010-10-31 18:01:20Z roland $
 * Copyright (C) 2007 Roland Krueger
 * Created on 03.12.2006
 *
 * Author: Roland Krueger (www.rolandkrueger.info)
 *
 * This file is part of RoKlib.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
package info.rolandkrueger.roklib.util.resources;

import info.rolandkrueger.roklib.util.helper.CheckForNull;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

public class ResourceHandlingManager
{
  protected static final String DEFAULT_LOCATION = ".";

  private static ResourceHandlingManager sSingletonInstance;

  private Map<ResourceHandle, URL> mResourceURLs;
  private Map<Class<?>, Set<String>> mMissingResources;
  protected String mResourceLocation = DEFAULT_LOCATION;

  private ResourceHandlingManager ()
  {
    mResourceURLs = new Hashtable<ResourceHandle, URL> ();
    mMissingResources = new Hashtable<Class<?>, Set<String>> ();
  }

  /**
   * Provides the singleton instance of the resource manager.
   * 
   * @return the manager's singleton instance
   */
  public synchronized static ResourceHandlingManager instance ()
  {
    if (sSingletonInstance == null) sSingletonInstance = new ResourceHandlingManager ();
    return sSingletonInstance;
  }

  private void addMissingResource (Class<?> clazz, String resourceName)
  {
    Set<String> set = mMissingResources.get (clazz);
    if (set == null)
    {
      set = Collections.synchronizedSet (new HashSet<String> ());
      mMissingResources.put (clazz, set);
    }
    set.add (resourceName);
  }

  public synchronized void addMissingResource (String resourceName)
  {
    addMissingResource (Class.class, resourceName);
  }

  public boolean areResourcesMissing ()
  {
    return mMissingResources.size () > 0;
  }

  public synchronized ResourceHandle registerResource (Class<?> referenceClass,
      String resourceName, boolean interpretNameAsURL) throws IOException
  {
    CheckForNull.check (referenceClass, resourceName);
    if (resourceName.equals (""))
      throw new IllegalArgumentException ("Resource name must not be empty.");

    String klassPath = referenceClass.getName ().replace ('.', '/') + ".class";
    String resourcePath = mResourceLocation + "/" + resourceName;
    String resourceIdentifier = String.format ("%s/%s", referenceClass.getName (), resourcePath);

    ClassLoader classloader = referenceClass.getClassLoader ();
    Enumeration<URL> resourceUrls = classloader.getResources (resourcePath);
    Enumeration<URL> classUrls = classloader.getResources (klassPath);

    URL classUrl;
    if (classUrls.hasMoreElements ())
    {
      classUrl = classUrls.nextElement ();
    } else
    {
      // didn't find the given class in the classpath
      throw new IllegalArgumentException (
          "The given class argument was not found in the classpath.");
    }

    String prefixToLookFor = classUrl.toString ().substring (0,
        classUrl.toString ().indexOf (klassPath));

    URL result = null;
    ResourceHandle resultHandle = new ResourceHandle (resourceIdentifier, referenceClass);
    while (resourceUrls.hasMoreElements ())
    {
      result = resourceUrls.nextElement ();
      if (result.toString ().startsWith (prefixToLookFor)) break;

      result = null;
    }
    if (result == null)
    {
      // resource was not found
      if (! interpretNameAsURL)
      {
        addMissingResource (referenceClass, referenceClass.getName () + "/" + resourcePath);
        return null;
      }
    } else
    {
      mResourceURLs.put (resultHandle, result);
    }

    if (interpretNameAsURL && getResourceURL (resultHandle) == null)
    {
      try
      {
        URL alternativeURL = new URL (resourceName);
        registerResource (alternativeURL, resourceName);
      } catch (MalformedURLException muExc)
      {
        return null;
      }
    }

    return resultHandle;
  }

  public synchronized ResourceHandle registerResource (Class<?> referenceClass, String resourceName)
      throws IOException
  {
    return registerResource (referenceClass, resourceName, false);
  }

  public synchronized ResourceHandle registerResource (URL urlToResource, String resourceIdentifier)
  {
    CheckForNull.check (urlToResource, resourceIdentifier);
    if (mResourceURLs.containsKey (resourceIdentifier))
    {
      throw new IllegalArgumentException ("A resource with identifier '" + resourceIdentifier
          + "' is already registered.");
    }
    ResourceHandle result = new ResourceHandle (resourceIdentifier, ResourceHandlingManager.class);
    mResourceURLs.put (result, urlToResource);
    return result;
  }

  public boolean isResourceRegisteredFor (String resourceIdentifier)
  {
    return mResourceURLs.containsKey (resourceIdentifier);
  }

  public InputStream getResourceData (ResourceHandle resourceHandle) throws IOException
  {
    URL url = mResourceURLs.get (resourceHandle);
    if (url == null) return null;
    return url.openStream ();
  }

  public InputStream getFileDataFallbackToResource (File resourceFile,
      ResourceHandle fallbackResource) throws IOException
  {
    if (resourceFile.exists () && resourceFile.canRead ())
      return new FileInputStream (resourceFile);
    return getResourceData (fallbackResource);
  }

  public URL getResourceURL (ResourceHandle handle)
  {
    return mResourceURLs.get (handle);
  }

  public List<String> getMissingResources ()
  {
    List<String> result = new Vector<String> ();

    for (Set<String> set : mMissingResources.values ())
    {
      result.addAll (set);
    }

    return result;
  }

  public String getMissingResourcesAsString ()
  {
    StringBuffer buf = new StringBuffer ();
    for (String s : getMissingResources ())
    {
      buf.append (s).append ("\n");
    }
    return buf.toString ();
  }

  public boolean resourcesMissingForClass (Class<?> clazz)
  {
    Set<String> set = mMissingResources.get (clazz);
    if (set == null) return false;
    return ! set.isEmpty ();
  }

  public void setResourceLocation (String location)
  {
    mResourceLocation = location;
  }

  public String getResourceLocation ()
  {
    return mResourceLocation;
  }

  public synchronized void reset ()
  {
    mMissingResources.clear ();
    mResourceURLs.clear ();
    mResourceLocation = DEFAULT_LOCATION;
  }
}
