/*
 * $Id: AbstractURLActionHandler.java 216 2010-12-30 12:39:20Z roland $
 * Copyright (C) 2007 - 2010 Roland Krueger
 * Created on 11.02.2010
 *
 * Author: Roland Krueger (www.rolandkrueger.info)
 *
 * This file is part of RoKlib.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
package info.rolandkrueger.roklib.webapps.urldispatching;

import info.rolandkrueger.roklib.util.LoggingManager;
import info.rolandkrueger.roklib.util.conditionalengine.AbstractCondition;
import info.rolandkrueger.roklib.util.helper.CheckForNull;
import info.rolandkrueger.roklib.util.net.IURLProvider;
import info.rolandkrueger.roklib.webapps.urldispatching.urlparameters.EnumURLParameterErrors;
import info.rolandkrueger.roklib.webapps.urldispatching.urlparameters.IURLParameter;

import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.Logger;

public abstract class AbstractURLActionHandler implements IURLActionHandler
{
  private static final long serialVersionUID = 8450975393827044559L;
  private static final String[] STRING_ARRAY_PROTOTYPE = new String[] {};
  private static final Logger LOG = LoggingManager.getInstance ().getLogger (AbstractURLActionHandler.class);
  
  private List<CommandForCondition> mCommandsForCondition;
  private List<IURLParameter<?>> mURLParameters;
  private List<String> mActionArgumentOrder;
  protected List<IURLActionHandler> mHandlerChain;
  private Map<String, AbstractURLActionHandler> mSubHandlers;
  private Map<String, List<Serializable>> mActionArgumentMap;
  private AbstractURLActionHandler mParentHandler;
  private AbstractURLActionCommand mDefaultCommand;
  protected String mActionName;
  private String mActionURI;
  private boolean mCaseSensitive = false;
  
  public AbstractURLActionHandler (String actionName)
  {
    CheckForNull.check (actionName);
    mActionName = actionName;
    mActionURI  = actionName;
    mDefaultCommand = null;
  }
  
  protected void setCaseSensitive (boolean caseSensitive)
  {
    mCaseSensitive = caseSensitive;
    if (mCaseSensitive & mSubHandlers != null)
    {
      Map<String, AbstractURLActionHandler> subHandlers = mSubHandlers;
      mSubHandlers = new HashMap<String, AbstractURLActionHandler> (4);
      mSubHandlers.putAll (subHandlers);
    }
  }
  
  private Map<String, AbstractURLActionHandler> getSubHandlerMap ()
  {
    if (mSubHandlers == null)
    {
      if (! mCaseSensitive)
      {
        mSubHandlers = new TreeMap<String, AbstractURLActionHandler> (String.CASE_INSENSITIVE_ORDER);
      }
      else
      {
        mSubHandlers = new HashMap<String, AbstractURLActionHandler> (4);
      }
    }
    return mSubHandlers;
  }
  
  public boolean isCaseSensitive ()
  {
    return mCaseSensitive;
  }

  public String getActionName ()
  {
    return mActionName;
  }
  
  public void setDefaultActionCommand (AbstractURLActionCommand command)
  {
    mDefaultCommand = command;
  }
  
  protected void registerURLParameter (IURLParameter<?> parameter)
  {
    if (parameter == null) return;
    if (mURLParameters == null)
      mURLParameters = new LinkedList<IURLParameter<?>> ();
    if ( ! mURLParameters.contains (parameter))
      mURLParameters.add (parameter);
  }
  
  protected void registerURLParameter (IURLParameter<?> parameter, boolean isOptional)
  {
    registerURLParameter (parameter);
    parameter.setOptional (isOptional);
  }
  
  protected boolean haveRegisteredURLParametersErrors ()
  {
    if (mURLParameters == null) return false;
    boolean result = false;
    
    for (IURLParameter<?> parameter : mURLParameters)
    {
      result |= parameter.getError () != EnumURLParameterErrors.NO_ERROR;
    }
    
    return result;
  }
  
  public final AbstractURLActionCommand handleURL (List<String> pUriTokens,
      Map<String, List<String>> pParameters, ParameterMode pParameterMode)
  {
    if (mCommandsForCondition != null)
    {
      for (CommandForCondition cfc : mCommandsForCondition)
      {
        if (cfc.mCondition.getBooleanValue () == true)
          return cfc.mDefaultCommandForCondition;
      }
    }
    if (mURLParameters != null)
    {
      if (pParameterMode == ParameterMode.QUERY)
      {
        for (IURLParameter<?> parameter : mURLParameters)
        {
          parameter.clearValue ();
          parameter.consume (pParameters);
        }
      } else
      {
        List<String> parameterNames = new LinkedList <String> ();
        for (IURLParameter<?> parameter : mURLParameters)
        {
          parameterNames.addAll (parameter.getParameterNames ());
        }
        if (pParameterMode == ParameterMode.DIRECTORY_WITH_NAMES)
        {
          Map<String, List<String>> parameters = new HashMap<String, List<String>> (4);
          String parameterName;
          String value;
          for (Iterator<String> it = pUriTokens.iterator (); it.hasNext ();)
          {
            parameterName = it.next ();
            try
            {
              parameterName = URLDecoder.decode (parameterName, "UTF-8");
            } catch (UnsupportedEncodingException e)
            {
              // nothing to do, parameterName stays encoded
            }
            value = "";
            if (parameterNames.contains (parameterName))
            {
              it.remove ();
              if (it.hasNext ())
              {
                value = it.next ();
                try
                {
                  value = URLDecoder.decode (value, "UTF-8");
                } catch (UnsupportedEncodingException e)
                {
                  // nothing to do, value stays encoded
                }
                it.remove ();
              }
              List<String> values = parameters.get (parameterName);
              if (values == null)
              {
                values = new LinkedList<String> ();
                parameters.put (parameterName, values);
              }
              values.add (value);
            }
          }
          for (IURLParameter<?> parameter : mURLParameters)
          {
            parameter.clearValue ();
            parameter.consume (parameters);
          }
        } else
        {
          List<String> valueList = new LinkedList<String> ();
          for (IURLParameter<?> parameter : mURLParameters)
          {
            parameter.clearValue ();
            if (pUriTokens.isEmpty ()) continue;
            valueList.clear ();
            int singleValueCount = parameter.getSingleValueCount ();
            int i = 0;
            while ( ! pUriTokens.isEmpty () && i < singleValueCount)
            {
              String token = pUriTokens.remove (0);
              try
              {
                token = URLDecoder.decode (token, "UTF-8");
              } catch (UnsupportedEncodingException e)
              {
                // nothing to do, token stays encoded
              }
              valueList.add (token);
              ++i;
            }
            parameter.consumeList (valueList.toArray (STRING_ARRAY_PROTOTYPE));
          }
        }
      }
    }

    if (mHandlerChain != null)
    {
      for (IURLActionHandler chainedHandler : mHandlerChain)
      {
        if (LOG.isTraceEnabled ())
        {
          LOG.trace ("Executing chained handler " + chainedHandler + " (" + mHandlerChain.size () + " chained handler(s) in list)");
        }
        AbstractURLActionCommand commandFromChain = chainedHandler.handleURL (pUriTokens, pParameters, pParameterMode);
        if (commandFromChain != null) return commandFromChain;
      }
    }
    
    return handleURLImpl (pUriTokens, pParameters, pParameterMode);
  }
      
  protected abstract AbstractURLActionCommand handleURLImpl (List<String> uriTokens, 
                                                             Map<String, List<String>> parameters, 
                                                             ParameterMode parameterMode);
  
  private String urlEncode (String term)
  {
    try
    {
      return URLEncoder.encode (term, "UTF-8");
    } catch (UnsupportedEncodingException e)
    {
      // this should not happen
      return term;
    }
  }
  
  public final void addSubHandler (AbstractURLActionHandler subHandler)
  {
    CheckForNull.check (subHandler);
    if (subHandler.mParentHandler != null)
      throw new IllegalArgumentException (String.format ("This sub-handler instance has " +
      		"already been added to another action handler. This handler = '%s'; sub-handler = '%s'", 
      		mActionName, subHandler.mActionName));
    subHandler.mParentHandler = this;
    subHandler.mActionURI = String.format ("%s%s%s", getActionURI (), "/", urlEncode (subHandler.mActionName));
    getSubHandlerMap ().put (subHandler.mActionName, subHandler);
    subHandler.setCaseSensitive (mCaseSensitive);
  }
  
  public String getActionURI ()
  {
    return mActionURI;
  }
  
  protected final void setParent (AbstractURLActionHandler parent)
  {
    mParentHandler = parent;
  }
  
  protected IURLProvider getContextURLProvider ()
  {
    if (mParentHandler != null)
    {
      return mParentHandler.getContextURLProvider ();
    } else
    {
      throw new IllegalStateException ("Unable to provide IURLProvider object: this URL " +
      		"action handler or one of its ancestors hasn't yet been added to a " + URLActionDispatcher.class.getName ());
    }
  }
  
  public URL getParameterizedActionURL (boolean clearAfterwards)
  {
    return getParameterizedActionURL (clearAfterwards, ParameterMode.QUERY);
  }
  
  public URL getParameterizedActionURL (boolean clearAfterwards, ParameterMode parameterMode)
  {
    return getParameterizedActionURL (clearAfterwards, parameterMode, false);
  }
  
  public URL getParameterizedActionURL (boolean clearAfterwards, ParameterMode parameterMode, boolean addHashMark)
  {
    return getParameterizedActionURL (clearAfterwards, parameterMode, addHashMark, null);
  }
  
  public URL getParameterizedActionURL (boolean clearAfterwards, ParameterMode parameterMode, boolean addHashMark, URL baseURL)
  {
    StringBuilder buf = new StringBuilder ();
    if (addHashMark) 
    {
      buf.append ('#');
      buf.append (getActionURI ().substring (1));
    } else
    {
      buf.append (getActionURI ());
    }
    
    boolean removeLastCharacter = false;
    if (mActionArgumentMap != null && ! mActionArgumentMap.isEmpty ())
    {
      if (parameterMode == ParameterMode.QUERY)
      {
        buf.append ('?');
        for (String argument : mActionArgumentOrder)
        {
          for (Serializable value : mActionArgumentMap.get (argument))
          {
            buf.append (urlEncode (argument)).append ('=').append (urlEncode (value.toString ()));
            buf.append ('&');
            removeLastCharacter = true;
          }
        }
      } else
      {
        buf.append ('/');
        for (String argument : mActionArgumentOrder)
        {
          for (Serializable value : mActionArgumentMap.get (argument))
          {
            if (parameterMode == ParameterMode.DIRECTORY_WITH_NAMES)
            {
              buf.append (urlEncode (argument)).append ('/');
            }
            buf.append (urlEncode (value.toString ()));
            buf.append ('/');
            removeLastCharacter = true;
          }
        }
      }
    }
    
    if (removeLastCharacter) buf.setLength (buf.length () - 1);
    
    try
    {
      URL url = baseURL == null ? getContextURLProvider ().getURL () : baseURL;
      return new URL (url, buf.toString ());
    } catch (MalformedURLException e)
    {
      throw new RuntimeException ("Unable to create URL object.", e);
    } finally
    {
      if (clearAfterwards)
      {
        clearActionArguments ();
      }
    }
  }
  
  public final void addDefaultCommandForCondition (AbstractURLActionCommand command, AbstractCondition condition)
  {
    CheckForNull.check (command, condition);
    if (mCommandsForCondition == null)  
      mCommandsForCondition = new LinkedList<CommandForCondition> ();
    CommandForCondition cfc = new CommandForCondition ();
    cfc.mDefaultCommandForCondition = command;
    cfc.mCondition = condition;
    mCommandsForCondition.add (cfc);
  }
  
  public void addToHandlerChain (IURLActionHandler handler)
  {
    CheckForNull.check (handler);
    if (mHandlerChain == null)
    {
      mHandlerChain = new LinkedList<IURLActionHandler> ();
    }
    mHandlerChain.add (handler);
  }
  
  /**
   * <code>null</code> argument values are ignored.
   * @param argumentName
   * @param argumentValues
   */
  public void addActionArgument (String argumentName, Serializable ... argumentValues)
  {
    CheckForNull.check (argumentName);
    if (mActionArgumentMap == null)
    {
      mActionArgumentMap = new HashMap<String, List<Serializable>> (4);
      mActionArgumentOrder = new LinkedList<String> ();
    }

    List<Serializable> valueList = mActionArgumentMap.get (argumentName);
    if (valueList == null)
    {
      valueList = new LinkedList<Serializable> ();
      mActionArgumentMap.put (argumentName, valueList);
    }
    for (Serializable value : argumentValues)
    {
      if (value != null)
        valueList.add (value);
    }
    if (valueList.isEmpty ())
    {
      mActionArgumentMap.remove (argumentName);
    } else if (!mActionArgumentOrder.contains (argumentName))
    {
      mActionArgumentOrder.add (argumentName);
    }
  }

  public void clearActionArguments ()
  {
    if (mActionArgumentMap != null)
    {
      mActionArgumentMap.clear ();
      mActionArgumentOrder.clear ();
    }
  }
  
  public final void clearDefaultCommands ()
  {
    mCommandsForCondition.clear ();
  }
  
  protected final AbstractURLActionCommand forwardToSubHandler (String handlerID, List<String> uriTokens, 
                                   Map<String, List<String>> parameters, ParameterMode parameterMode)
  {
    
    AbstractURLActionHandler subHandler = getSubHandlerMap ().get (handlerID);
    if (subHandler == null) return mDefaultCommand;
    
    return subHandler.handleURL (uriTokens, parameters, parameterMode);
  }
  
  private static class CommandForCondition implements Serializable
  {
    private static final long serialVersionUID = 2090692709855753816L;

    private AbstractURLActionCommand mDefaultCommandForCondition;
    private AbstractCondition        mCondition;
  }
  
  public void getActionURLOverview (List<String> targetList)
  {
    StringBuilder buf = new StringBuilder ();
    buf.append (getActionURI ());
    
    if (mURLParameters != null && mURLParameters.size () > 0)
    {
      buf.append (" ? ");
      for (IURLParameter<?> parameter : mURLParameters)
      {
        buf.append (parameter).append (", ");
      }
      buf.setLength (buf.length () - 2);
    }
    if (buf.length () > 0)
      targetList.add (buf.toString ());
    for (AbstractURLActionHandler subHandler : getSubHandlerMap ().values ())
    {
      subHandler.getActionURLOverview (targetList);
    }
  }
  
  @Override
  public String toString ()
  {
    return String.format ("%s='%s'", getClass ().getSimpleName (), mActionName);
  }
}
