/*
 * $Id: URLActionDispatcher.java 214 2010-12-30 11:32:45Z roland $
 * Copyright (C) 2007 - 2010 Roland Krueger
 * Created on 10.02.2010
 *
 * Author: Roland Krueger (www.rolandkrueger.info)
 *
 * This file is part of RoKlib.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */
package info.rolandkrueger.roklib.webapps.urldispatching;

import info.rolandkrueger.roklib.util.LoggingManager;
import info.rolandkrueger.roklib.util.helper.CheckForNull;
import info.rolandkrueger.roklib.util.net.IURLProvider;
import info.rolandkrueger.roklib.webapps.data.DownloadInfo;
import info.rolandkrueger.roklib.webapps.urldispatching.IURLActionHandler.ParameterMode;

import java.io.Serializable;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.Logger;

public class URLActionDispatcher implements Serializable
{
  private static final long serialVersionUID = 7151587763812706383L;
  private static final Logger LOG = LoggingManager.getInstance ().getLogger (URLActionDispatcher.class);
  
  private Map<String, List<String>>    mCurrentParameters;
  private URL                          mContextOriginal;
  private String                       mRelativeUriOriginal;
  private Map<String, String[]>        mCurrentParametersOriginalValues;
  private AbstractURLActionCommand     m404FileNotFoundCommand;
  private DispatchingURLActionHandler  mRootDispatcher;
  private IURLActionDispatcherListener mListener;
  private ParameterMode                mParameterMode = ParameterMode.QUERY;
  private IURLProvider                 mURLProvider;
  
  public URLActionDispatcher (IURLProvider urlProvider, boolean useCaseSensitiveURLs)
  {
    CheckForNull.check (urlProvider);
    mURLProvider = urlProvider;
   
    if (useCaseSensitiveURLs)
    {
      mCurrentParameters = new HashMap<String, List<String>> ();      
    } else
    {
      mCurrentParameters =  new TreeMap<String, List<String>> (String.CASE_INSENSITIVE_ORDER);
    }
    mRootDispatcher = new DispatchingURLActionHandler ("");
    mRootDispatcher.setCaseSensitive (useCaseSensitiveURLs);
    mRootDispatcher.setParent (new AbstractURLActionHandler ("")
    {
      private static final long serialVersionUID = 3744506992900879054L;

      @Override
      protected IURLProvider getContextURLProvider ()
      {
        return mURLProvider;
      }
      
      protected AbstractURLActionCommand handleURLImpl (List<String> uriTokens, Map<String, List<String>> parameters, 
          ParameterMode parameterMode)
      {
        return null;
      }
    });
  }  
  
  public boolean isCaseSensitive ()
  {
    return mRootDispatcher.isCaseSensitive ();
  }
  
  public DispatchingURLActionHandler getRootActionHandler ()
  {
    return mRootDispatcher;
  }
  
  public void setURLActionDispatcherListener (IURLActionDispatcherListener listener)
  {
    mListener = listener;
  }
  
  public void set404FileNotFoundCommand (AbstractURLActionCommand fileNotFoundCommand)
  {
    m404FileNotFoundCommand = fileNotFoundCommand;
  }
  
  protected void setParameters (Map<String, List<String>> parameters)
  {
    mCurrentParameters.clear ();
    mCurrentParameters.putAll (parameters);
  }
  
  protected Map<String, List<String>> getParameters ()
  {
    return mCurrentParameters;
  }
  
  public void clearParameters ()
  {
    mCurrentParameters.clear ();
  }
  
  public void handleParameters (Map<String, String[]> parameters)
  {
    if (parameters == null) return;
    mCurrentParameters.clear ();
    mCurrentParametersOriginalValues = parameters;
    for (String key : parameters.keySet ())
    {
      List<String> params = new ArrayList<String> (Arrays.asList (parameters.get (key)));
      if ( ! params.isEmpty ())
        mCurrentParameters.put (key, params);
    }
  }
    
  public void setParameterMode (ParameterMode parameterMode)
  {
    mParameterMode = parameterMode;
  }
  
  public DownloadInfo handleURIAction (URL context, String relativeUri)
  {
    return handleURIAction (context, relativeUri, mParameterMode);
  }
  
  public DownloadInfo handleURIAction (URL context, String relativeUri, ParameterMode parameterMode)
  {
    mContextOriginal = context;
    mRelativeUriOriginal = relativeUri;
    List<String> uriTokens = new ArrayList<String> (Arrays.asList (relativeUri.split ("/")));
    
    if (LOG != null && LOG.isTraceEnabled ())
      LOG.trace (String.format ("Dispatching URI: '%s', params: '%s'", relativeUri, mCurrentParameters));
    
    AbstractURLActionCommand action = mRootDispatcher.handleURL (uriTokens, mCurrentParameters, 
        parameterMode);
    if (action == null)
    {
      // execute 404 File Not Found command
      LOG.info ("404 File Not Found: " + relativeUri + "?" + mCurrentParameters);
      if (m404FileNotFoundCommand != null)
        m404FileNotFoundCommand.execute ();
      return null;
    } else
    {
      action.execute ();
      if (mListener != null) mListener.handleURLActionCommand (action);
      return action.getDownloadStream ();
    }
  }
  
  public String getCurrentlyHandledURI ()
  {
    return mRelativeUriOriginal;
  }
  
  public DownloadInfo replayCurrentAction ()
  {
    handleParameters (mCurrentParametersOriginalValues);
    return handleURIAction (mContextOriginal, mRelativeUriOriginal);
  }
}
